from unittest.mock import MagicMock
from aas_timeseries.data import Data
from aas_timeseries.layers import Markers


def test_tooltip_options():

    fig = MagicMock()
    time_series = MagicMock()
    time_series.colnames = ['the_time', 'the_flux', 'the_other']
    data = Data(time_series)

    marker = Markers(parent=fig)
    marker.data = data
    marker.time_column = 'the_time'
    marker.column = 'the_flux'

    tooltip = marker.to_vega()[0]['encode']['hover']['tooltip']
    assert tooltip == {'signal': "{'the_time': datum.the_time, 'the_flux': datum.the_flux}"}

    marker.tooltip = False

    assert 'hover' not in marker.to_vega()[0]['encode']

    marker.tooltip = ('the_time', 'the_other')
    tooltip = marker.to_vega()[0]['encode']['hover']['tooltip']
    assert tooltip == {'signal': "{'the_time': datum.the_time, 'the_other': datum.the_other}"}

    marker.tooltip = ['the_flux']
    tooltip = marker.to_vega()[0]['encode']['hover']['tooltip']
    assert tooltip == {'signal': "{'the_flux': datum.the_flux}"}

    marker.tooltip = {'the_time': 'Time', 'the_flux': 'Flux'}
    tooltip = marker.to_vega()[0]['encode']['hover']['tooltip']
    assert tooltip == {'signal': "{'Time': datum.the_time, 'Flux': datum.the_flux}"}



import os
from traceback import print_exc

import pytest

from matplotlib.testing.compare import compare_images

DATA = os.path.abspath(os.path.join(os.path.dirname(__file__), 'data'))


def compare_to_reference_json(tmpdir, test_name, image_tests=False):

    tmpdir = tmpdir.strpath

    expected_files = sorted(os.listdir(os.path.join(DATA, test_name)))
    actual_files = sorted(os.listdir(tmpdir))

    if not image_tests:
        expected_files = [x for x in expected_files if not x.endswith('.png')]

    assert expected_files == actual_files

    for filename in expected_files:

        expected_file = os.path.join(DATA, test_name, filename)
        actual_file = os.path.join(tmpdir, filename)

        if filename.endswith('.png'):

            try:
                msg = compare_images(expected_file, actual_file, tol=0)
            except Exception:
                msg = 'Image comparison failed:'
                print_exc()

            if msg is not None:
                pytest.fail(msg, pytrace=False)

        else:

            with open(expected_file) as f:
                expected = f.read().strip()
            with open(actual_file) as f:
                actual = f.read().strip()

            # Normalize line endings
            expected = expected.replace('\r\n', '\n').replace('\r', '\n')
            actual = actual.replace('\r\n', '\n').replace('\r', '\n')

            assert expected == actual



# -*- coding: utf-8 -*-

from matplotlib import pyplot as plt

import pytest

from aas_timeseries.matplotlib import (PhaseAsDegreesLocator,
                                       PhaseAsDegreesFormatter,
                                       PhaseAsRadiansLocator,
                                       PhaseAsRadiansFormatter)


def get_ticklabels(axis):
    return [x.get_text() for x in axis.get_ticklabels()]


DEG_CASES = [((-20, 50), ['-5000°', '0°', '5000°', '10000°', '15000°']),
             ((0., 1.), ['0°', '60°', '120°', '180°', '240°', '300°', '360°']),
             ((0.3, 0.4), ['108°', '114°', '120°', '126°', '132°', '138°', '144°']),
             ((0.555, 0.556), ['199.84°', '199.92°', '200°', '200.08°', '200.16°'])]


@pytest.mark.parametrize(('limits', 'expected'), DEG_CASES)
def test_phase_degrees(tmpdir, limits, expected):

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    ax.xaxis.set_major_locator(PhaseAsDegreesLocator())
    ax.xaxis.set_major_formatter(PhaseAsDegreesFormatter())
    ax.set_xlim(*limits)
    fig.savefig(tmpdir.join('figure.png'))
    assert get_ticklabels(ax.xaxis) == expected


RAD_CASES = [((-20, 50), ['-32π', '-16π', '0', '16π', '32π', '48π', '64π', '80π', '96π']),
             ((0., 1.), ['0', 'π/2', 'π', '3π/2', '2π']),
             ((0.3, 0.4), ['5π/8', '11π/16', '3π/4']),
             ((0.555, 0.556), ['1.1104π', '1.1108π', '1.1113π', '1.1118π'])]


@pytest.mark.parametrize(('limits', 'expected'), RAD_CASES)
def test_phase_radians(tmpdir, limits, expected):

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    ax.xaxis.set_major_locator(PhaseAsRadiansLocator())
    ax.xaxis.set_major_formatter(PhaseAsRadiansFormatter())
    ax.set_xlim(*limits)
    fig.savefig(tmpdir.join('figure.png'))
    assert get_ticklabels(ax.xaxis) == expected



import os
import tempfile
from io import StringIO
from json import dump
from zipfile import ZipFile

import numpy as np

from matplotlib import pyplot as plt

from astropy.time import Time
from astropy.table import Table
from astropy import units as u

from astropy.visualization import quantity_support

from aas_timeseries.backports import time_support
from aas_timeseries.colors import auto_assign_colors
from aas_timeseries.views import BaseView, View
from aas_timeseries.matplotlib import (PhaseAsDegreesLocator,
                                       PhaseAsDegreesFormatter,
                                       PhaseAsRadiansLocator,
                                       PhaseAsRadiansFormatter)

__all__ = ['InteractiveTimeSeriesFigure']


class InteractiveTimeSeriesFigure(BaseView):
    """
    An interactive time series figure.

    Parameters
    ----------
    width : int, optional
        The preferred width of the figure, in pixels.
    height : int, optional
        The preferred height of the figure, in pixels.
    padding : float, optional
        The padding inside the axes, as a fraction of the size of the axes
    resize : bool, optional
        Whether to resize the figure to the available space.
    title : str, optinal
        If views are added to the figure, this title is used for the default
        view, otherwise 'Default' is used.
    """

    def __init__(self, width=600, height=400, padding=36, resize=False, title=None, time_mode=None):
        super().__init__(time_mode=time_mode)
        self._width = width
        self._height = height
        self._resize = resize
        self._padding = padding
        self._yunit = 'auto'
        self._views = []
        self._title = title

    @property
    def yunit(self):
        """
        The unit to use for the y axis, or 'auto' to guess. If using data
        without quantities, this can be set to ``u.one``.
        """
        return self._yunit

    @yunit.setter
    def yunit(self, value):
        if isinstance(value, u.UnitBase) or value == 'auto':
            self._yunit = value
        else:
            self._yunit = u.Unit(value)

    def _guess_yunit(self):
        """
        Try and guess unit to use on the y axis based on the limits set and the
        data used.
        """

        # First check if the limits were set explicitly and if so use those
        # units if they were Quantities since it seems likely this was
        # deliberate.
        if self.ylim is not None and isinstance(self.ylim[0], u.Quantity):
            return self.ylim[0].unit

        # Next, check through all the data required by the layers, first in the
        # figure, then in the views, and return the first unit or lack of unit
        # found.
        for layer in self._layers:
            for (data, colname) in layer._required_ydata:
                return data.unit(colname)
        for view in self._views:
            for layer in view['view']._layers:
                for (data, colname) in layer._required_ydata:
                    return data.unit(colname)

        # Since we didn't find anything, let's assume that the y axis should
        # be dimensionless. In theory we could also go through non-data layers
        # and check the units used there.
        return u.one

    def add_view(self, title, description=None, include=None, exclude=None, empty=False, time_mode=None):

        if empty:
            inherited_layers = {}
        elif include is not None:
            for layer in include:
                if layer not in self._layers:
                    raise ValueError(f'Layer {layer} does not exist in base figure')
            inherited_layers = {layer: self._layers[layer] for layer in include}
        elif exclude is not None:
            for layer in exclude:
                if layer not in self._layers:
                    raise ValueError(f'Layer {layer} does not exist in base figure')
            inherited_layers = {layer: self._layers[layer] for layer in self._layers if layer not in exclude}
        else:
            inherited_layers = self._layers.copy()

        view = View(figure=self, inherited_layers=inherited_layers, time_mode=time_mode)

        self._views.append({'title': title, 'description': description, 'view': view})

        return view

    def export_interactive_bundle(self, filename, embed_data=False,
                                  minimize_data=True, override_style=False):
        """
        Create a bundle for the interactive figure containing an HTML file and
        JSON file along with any required CSV files.

        Parameters
        ----------
        filename : str
            The filename for the zip file
        embed_data : bool, optional
            Whether to embed the data in the JSON file (`True`) or include it
            in separate CSV files (`False`). The default is `False`.
        minimize_data : bool, optional
            Whether to include only data required for the visualization (`True`)
            or also other unused columns/fields in the time series (`False`).
            The default is `True`.
        override_style : bool, optional
            By default, any unspecified colors will be automatically chosen.
            If this parameter is set to `True`, all colors will be reassigned,
            even if already set.
        """

        start_dir = os.path.abspath('.')
        tmp_dir = tempfile.mkdtemp()
        os.chdir(tmp_dir)
        try:
            self.save_vega_json('figure.json', embed_data=embed_data)
        finally:
            os.chdir(start_dir)
        html_file = os.path.join(os.path.dirname(__file__), 'screenshot', 'template.html')
        with ZipFile(filename, 'w') as fzip:
            for filename in os.listdir(tmp_dir):
                fzip.write(os.path.join(tmp_dir, filename), os.path.basename(filename))
            fzip.write(html_file, 'index.html')

    def _check_colors(self, override_style=False):
        # Auto-assign colors if needed
        colors = auto_assign_colors(self._layers)
        for layer, color in zip(self._layers, colors):
            if override_style or layer.color is None:
                layer.color = color

    def save_static(self, prefix, format='png', override_style=False):
        """
        Export the figure to one or more static files using Matplotlib. If views
        are present then one plot is produced for each view.

        Parameters
        ----------
        prefix : str
            The name of the plot (without extension).
        format : str
            Any valid format supported by Matplotlib.
        override_style : bool, optional
            By default, any unspecified colors will be automatically chosen.
            If this parameter is set to `True`, all colors will be reassigned,
            even if already set.
        """

        # Start off by figuring out what units we are using on the y axis.
        # Note that we check the consistency of the units only here for
        # simplicity otherwise any guessing while users add/remove layers is
        # tricky.
        yunit = self._guess_yunit() if self.yunit == 'auto' else self.yunit

        # Auto-assign colors if needed
        self._check_colors(override_style=override_style)

        # We now loop over the main figure and all the views, and produce a
        # static plot for each of them.

        def pad_limits(limits, padding):
            vrange = (limits[1] - limits[0]) * padding
            return limits[0] - vrange, limits[1] + vrange

        for iview, view in enumerate([self] + self._views):

            if view is not self:
                view = view['view']

            # Note that if we aren't dealing with non-absolute times, the
            # following settings don't matter since the time_support context
            # manager won't have any effect.

            if view.time_format == 'auto' or view.time_mode != 'absolute':
                time_format = 'iso'
                simplify = True
            else:
                time_format = view.time_format
                simplify = False

            with time_support(format=time_format, simplify=simplify, scale='utc'):
                with quantity_support():

                    fig = plt.figure(figsize=(self._width / 100,
                                              self._height / 100))
                    ax = fig.add_axes([0.15, 0.12, 0.8, 0.86])

                    for layer in view.layers:
                        layer.to_mpl(ax, yunit=yunit)

            if view.time_mode == 'phase':
                if view.time_format == 'degrees':
                    ax.xaxis.set_major_locator(PhaseAsDegreesLocator())
                    ax.xaxis.set_major_formatter(PhaseAsDegreesFormatter())
                elif view.time_format == 'radians':
                    ax.xaxis.set_major_locator(PhaseAsRadiansLocator())
                    ax.xaxis.set_major_formatter(PhaseAsRadiansFormatter())

            x_domain, y_domain = view._get_domains(yunit, as_vega=False)

            ax.set_xlim(*x_domain)
            ax.set_ylim(*y_domain)

            # Apply padding - we just get the limits again because the x limits
            # above may have been Time objects, so we get the limits again from
            # Matplotlib.
            ax.set_xlim(*pad_limits(ax.get_xlim(), self._padding / self._width))
            ax.set_ylim(*pad_limits(ax.get_ylim(), self._padding / self._height))

            if view is self:
                filename = prefix + '.' + format
            else:
                filename = prefix + '_view' + str(iview) + '.' + format

            ax.set_xlabel(view.xlabel)
            ax.set_ylabel(view.ylabel)

            fig.savefig(filename)

    def save_vega_json(self, filename, embed_data=False,
                       minimize_data=True, override_style=False):
        """
        Export the JSON file, and optionally CSV data files.

        Parameters
        ----------
        filename : str
            The name of the JSON file
        embed_data : bool, optional
            Whether to embed the data in the JSON file (`True`) or include it
            in separate CSV files (`False`). The default is `False`.
        minimize_data : bool, optional
            Whether to include only data required for the visualization (`True`)
            or also other unused columns/fields in the time series (`False`).
            The default is `True`.
        override_style : bool, optional
            By default, any unspecified colors will be automatically chosen.
            If this parameter is set to `True`, all colors will be reassigned,
            even if already set.
        """

        # Start off by figuring out what units we are using on the y axis.
        # Note that we check the consistency of the units only here for
        # simplicity otherwise any guessing while users add/remove layers is
        # tricky.
        yunit = self._guess_yunit() if self.yunit == 'auto' else self.yunit

        # Auto-assign colors if needed
        self._check_colors(override_style=override_style)

        # Start off with empty JSON
        json = {}

        # Schema
        json['$schema'] = 'https://vega.github.io/schema/vega/v4.json'

        # Layout
        json['title'] = self._title or 'Default'
        json['width'] = self._width
        json['height'] = self._height
        json['padding'] = 0
        json['autosize'] = {'type': 'fit', 'resize': self._resize}

        json['_extend'] = {}

        # Data

        # We start off by checking which columns and data are going to be
        # required. We do this by iterating over the layers in the main
        # figure and the views and keeping track of the set of (data, column)
        # that are needed.
        required_xdata = []
        required_ydata = []
        required_tooltipdata = []
        for layer in self._layers:
            required_xdata.extend(layer._required_xdata)
            required_ydata.extend(layer._required_ydata)
            required_tooltipdata.extend(layer._required_tooltipdata)
        for view in self._views:
            for layer in view['view']._layers:
                required_xdata.extend(layer._required_xdata)
                required_ydata.extend(layer._required_ydata)
                required_tooltipdata.extend(layer._required_tooltipdata)
        required_xdata = set(required_xdata)
        required_ydata = set(required_ydata)
        required_tooltipdata = set(required_tooltipdata)

        json['data'] = []

        for data in self._data.values():

            # Start off by constructing a new table with only the subset of
            # columns required, and the time as an ISO string. Note that we
            # need to explicitly specify that we want UTC times, then add the
            # Z suffix since this isn't something that astropy does. For
            # relative times we always use seconds, and for phases we use values
            # in the range [0:1].
            table = Table()
            time_columns = []
            for colname in data.time_series.colnames:
                if (not minimize_data or (data, colname) in required_xdata | required_ydata | required_tooltipdata):
                    column = data.time_series[colname]
                    if isinstance(column, Time):
                        table[colname] = np.char.add(column.utc.isot, 'Z')
                        time_columns.append(colname)
                    elif (data, colname) in required_xdata:
                        try:
                            table[colname] = data.column_to_values(colname, u.s)
                        except u.UnitsError:
                            table[colname] = data.column_to_values(colname, u.one)
                    elif (data, colname) in required_ydata:
                        table[colname] = data.column_to_values(colname, yunit)
                    else:
                        table[colname] = column

            # Next up, we create the information for the 'parse' Vega key
            # which indicates the format for each column.
            parse = {}
            for colname in table.colnames:
                column = table[colname]
                if colname in time_columns:
                    parse[colname] = 'date'
                elif column.dtype.kind in 'fi':
                    parse[colname] = 'number'
                elif column.dtype.kind in 'b':
                    parse[colname] = 'boolean'
                else:
                    parse[colname] = 'string'

            vega = {'name': data.uuid,
                    'format': {'type': 'csv',
                               'parse': parse}}

            # We now either embed the CSV inside the JSON or create CSV files.
            # For now we use UUIDs for the data file names in the latter case
            # but in future we could find a way to preserve information about
            # the original filenames the data came from.
            if embed_data:
                s = StringIO()
                table.write(s, format='ascii.basic', delimiter=',')
                s.seek(0)
                # NOTE: when embedding the data inside the JSON file, we should
                # just use simple Unix line endings inside the serialized table.
                csv_string = s.read().replace('\r\n', '\n')
                vega['values'] = csv_string
            else:
                data_filename = 'data_' + data.uuid + '.csv'
                data_path = os.path.join(os.path.dirname(filename), data_filename)
                table.write(data_path, format='ascii.basic', delimiter=',')
                vega['url'] = data_filename

            json['data'].append(vega)

        # At this point, we loop over all the views (including the main view
        # given by self) and output these to the JSON.

        for view in [self] + self._views:

            if view is self:

                view_json = json

            else:

                view_json = {'name': self.uuid,
                             'title': view['title'],
                             'description': view['description']}

                if '_views' not in json:
                    json['_views'] = []

                json['_views'].append(view_json)

                view = view['view']

            if view._time_mode == 'absolute':
                x_type = 'time'
                x_input = 'iso'
            elif view._time_mode == 'relative':
                x_type = 'number'
                x_input = 'seconds'
            elif view._time_mode == 'phase':
                x_type = 'number'
                x_input = 'unity'

            view_json['_extend'] = {'scales': [{'name': 'xscale',
                                                'input': x_input,
                                                'output': view.time_format}]}

            view_json['axes'] = [{'orient': 'bottom',
                                  'scale': 'xscale',
                                  'title': view.xlabel},
                                 {'orient': 'left',
                                  'scale': 'yscale',
                                  'title': view.ylabel}]

            view_json['scales'] = [{'name': 'xscale',
                                    'type': x_type,
                                    'range': 'width',
                                    'zero': False,
                                    'padding': self._padding},
                                   {'name': 'yscale',
                                    'type': 'log' if view.ylog else 'linear',
                                    'range': 'height',
                                    'zero': False,
                                    'padding': self._padding}]

            # Limits, if specified

            x_domain, y_domain = view._get_domains(yunit)

            if x_domain is not None:
                view_json['scales'][0]['domain'] = x_domain

            if y_domain is not None:
                view_json['scales'][1]['domain'] = y_domain

            if view is self:

                view_json['marks'] = []
                for layer, settings in self._layers.items():
                    view_json['marks'].extend(layer.to_vega(yunit=yunit))

            else:

                view_json['markers'] = []

                for layer, settings in view._inherited_layers.items():
                    for uuid in layer.uuids:
                        view_json['markers'].append({'name': uuid,
                                                     'visible': settings['visible']})

                for layer, settings in view._layers.items():

                    if 'marks' not in json['_extend']:
                        json['_extend']['marks'] = []

                    json['_extend']['marks'].extend(layer.to_vega(yunit=yunit))
                    for uuid in layer.uuids:
                        view_json['markers'].append({'name': uuid,
                                                     'visible': settings['visible']})

        with open(filename, 'w') as f:
            dump(json, f, indent='  ', sort_keys=True)

    def preview_interactive(self):
        """
        Show an interactive version of the figure (only works in Jupyter
        notebook or lab).
        """
        # FIXME: should be able to do without a file
        from jupyter_aas_timeseries import TimeSeriesWidget
        tmpfile = tempfile.mktemp()
        self.save_vega_json(tmpfile, embed_data=True, minimize_data=True)
        widget = TimeSeriesWidget(tmpfile)
        return widget

    def remove(self, layer):
        """
        Remove a layer from the figure.
        """
        if layer in self._layers:
            self._layers.pop(layer)
            for view in self._views:
                if layer in view['view'].layers:
                    view['view'].remove(layer)
        else:
            raise ValueError(f"Layer '{layer.label}' is not in figure")



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from pkg_resources import get_distribution, DistributionNotFound

__all__ = ['InteractiveTimeSeriesFigure']

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    pass

from .visualization import InteractiveTimeSeriesFigure



# The code in this module has been adapted from the PyWWT package which was
# originally released under the following license:
#
# Copyright (c) 2017, Thomas P. Robitaille, O. Justin Otor, and John ZuHone
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.
# * Neither the name of the WorldWide Telescope project nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np

from traitlets import (TraitType, TraitError,
                       Any as OriginalAny,
                       Bool as OriginalBool,
                       CFloat as OriginalCFloat,
                       Int as OriginalInt,
                       Unicode as OriginalUnicode)

from astropy import units as u
from astropy.time import Time

try:
    from matplotlib.colors import to_hex
except ImportError:
    from matplotlib.colors import colorConverter, rgb2hex
    def to_hex(input):  # noqa
        return rgb2hex(colorConverter.to_rgb(input))

from aas_timeseries.data import Data

__all__ = ['Any', 'Bool', 'CFloat', 'PositiveCFloat', 'Int', 'Unicode',
           'UnicodeChoice', 'AstropyQuantity', 'DataTrait', 'ColumnTrait',
           'AstropyTime', 'Color', 'Opacity', 'Tooltip']


# We inherit the original trait classes to make sure that the docstrings are set


class Any(OriginalAny):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.help:
            self.__doc__ = self.help


class Bool(OriginalBool):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.help:
            self.__doc__ = self.help


class CFloat(OriginalCFloat):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.help:
            self.__doc__ = self.help


class PositiveCFloat(CFloat):

    def validate(self, obj, value):
        if value >= 0:
            return value
        else:
            raise TraitError('value should be positive')


class Int(OriginalInt):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.help:
            self.__doc__ = self.help


class Unicode(OriginalUnicode):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.help:
            self.__doc__ = self.help


class UnicodeChoice(Unicode):

    def __init__(self, *args, choices=[], **kwargs):
        super().__init__(*args, **kwargs)
        self.choices = choices
        choices_doc = []
        for choice in self.choices:
            if choice == self.default_value:
                choices_doc.append(f"``'{choice}'`` (default)")
            else:
                choices_doc.append(f"``'{choice}'``")
        self.__doc__ += '\n\nOne of ' + (', '.join(choices_doc)) + '.'

    def validate(self, obj, value):
        if value in self.choices:
            return value
        else:
            raise TraitError('value should be one of ' + ', '.join(f"'{choice}'" for choice in self.choices))


class AstropyQuantity(TraitType):

    default = 0 * u.one
    info_text = '\'Custom trait to handle astropy quantities with units\''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.help:
            self.__doc__ = self.help

    def validate(self, obj, value):
        if isinstance(value, u.Quantity):
            return value
        elif np.isscalar(value) and np.isreal(value):
            return value * u.one
        self.error(obj, value)


class DataTrait(TraitType):

    default = None
    info_text = "'Custom trait to handle Data objects'"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.help:
            self.__doc__ = self.help

    def validate(self, obj, value):
        if isinstance(value, Data):
            return value
        else:
            raise TraitError('value should be a Data instance')


class ColumnTrait(Unicode):

    allow_none = True
    default = None
    info_text = "'Custom trait to handle columns of Data objects'"

    def validate(self, obj, value):
        if value is None:
            return value
        try:
            obj.data
        except TraitError:
            raise TraitError('data should be set before column')
        if value in obj.data.time_series.colnames:
            return value
        else:
            raise TraitError(f'{value} is not a valid column name')


class AstropyTime(TraitType):

    default = None
    info_text = "'Custom trait to handle astropy time'"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.help:
            self.__doc__ = self.help

    def validate(self, obj, value):
        if isinstance(value, Time):
            return value
        else:
            try:
                return Time(value)
            except ValueError:
                raise TraitError('value should be a Time instance')


class Color(TraitType):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.help:
            self.__doc__ = self.help

    def validate(self, obj, value):
        if value is None:
            return None
        elif isinstance(value, str) or (isinstance(value, tuple) and len(value) == 3):
            return to_hex(value)
        else:
            if hasattr(obj, 'opacity'):
                raise TraitError('color must be a string or a tuple of 3 or 4 floats')
            else:
                raise TraitError('color must be a string or a tuple of 3 floats')


class Opacity(OriginalCFloat):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.help:
            self.__doc__ = self.help

    def validate(self, obj, value):
        if 0 <= value <= 1:
            return value
        else:
            raise TraitError("opacity must be a value in the range [0:1]")


class Tooltip(TraitType):

    default = False
    info_text = "'Whether or not to show a tooltip, and optionally a list or dictionary of values to show'"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.help:
            self.__doc__ = self.help

    def validate(self, obj, value):
        if isinstance(value, (bool, tuple, list, dict)):
            return value
        else:
            raise TraitError('value should be a boolean, tuple, list, or dictionary')



# Backport from Astropy - can be removed once Astropy 4.0 or later is required.
#
# Copyright (c) 2011-2017, Astropy Developers
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.
# * Neither the name of the Astropy Team nor the names of its contributors may be
#   used to endorse or promote products derived from this software without
#   specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import re

import numpy as np

from datetime import datetime

from astropy.time import Time
from astropy import units as u
from astropy.utils.exceptions import ErfaError

__all__ = ['time_support']

__doctest_skip__ = ['*']

UNSUPPORTED_FORMATS = ('datetime', 'datetime64')
YMDHMS_FORMATS = ('fits', 'iso', 'isot', 'yday')
STR_FORMATS = YMDHMS_FORMATS + ('byear_str', 'jyear_str')


def time_support(*, scale=None, format=None, simplify=True):
    """
    Enable support for plotting `astropy.time.Time` instances in
    matplotlib.

    May be (optionally) used with a ``with`` statement.

      >>> import matplotlib.pyplot as plt
      >>> from astropy import units as u
      >>> from astropy import visualization
      >>> with visualization.time_support():  # doctest: +IGNORE_OUTPUT
      ...     plt.figure()
      ...     plt.plot(Time(['2016-03-22T12:30:31', '2016-03-22T12:30:38', '2016-03-22T12:34:40']))
      ...     plt.draw()

    Parameters
    ----------
    scale : str, optional
        The time scale to use for the times on the axis. If not specified,
        the scale of the first Time object passed to Matplotlib is used.
    format : str, optional
        The time format to use for the times on the axis. If not specified,
        the format of the first Time object passed to Matplotlib is used.
    simplify : bool, optional
        If possible, simplify labels, e.g. by removing 00:00:00.000 times from
        ISO strings if all labels fall on that time, or by removing the leading
        date when it is repeated.
    """

    import matplotlib.units as units
    from matplotlib.ticker import MaxNLocator, ScalarFormatter
    from astropy.visualization.wcsaxes.utils import select_step_hour, select_step_scalar

    class AstropyTimeLocator(MaxNLocator):

        # Note: we default to AutoLocator since many time formats
        # can just use this.

        def __init__(self, converter, *args, **kwargs):
            kwargs['nbins'] = 4
            super().__init__(*args, **kwargs)
            self._converter = converter

        def tick_values(self, vmin, vmax):

            # Where we put the ticks depends on the format we are using
            if self._converter.format in YMDHMS_FORMATS:

                # If we are here, we need to check what the range of values
                # is and decide how to find tick locations accordingly

                vrange = vmax - vmin

                if (self._converter.format != 'yday' and vrange > 31) or vrange > 366:  # greater than a month

                    # We need to be careful here since not all years and months have
                    # the same length

                    # Start off by converting the values from the range to
                    # datetime objects, so that we can easily extract the year and
                    # month.

                    try:
                        tmin = Time(vmin, scale=self._converter.scale, format='mjd').datetime
                        tmax = Time(vmax, scale=self._converter.scale, format='mjd').datetime
                    except ErfaError:
                        return []

                    # Find the range of years
                    ymin = tmin.year
                    ymax = tmax.year

                    if ymax > ymin + 1:  # greater than a year

                        # Find the step we want to use
                        ystep = int(select_step_scalar(max(1, (ymax - ymin) / 3)))

                        ymin = ystep * (ymin // ystep)

                        # Generate the years for these steps
                        times = []
                        for year in range(ymin, ymax + 1, ystep):
                            times.append(datetime(year=year, month=1, day=1))

                    else:  # greater than a month but less than a year

                        mmin = tmin.month
                        mmax = tmax.month + 12 * (ymax - ymin)

                        mstep = int(select_step_scalar(max(1, (mmax - mmin) / 3)))

                        mmin = mstep * max(1, mmin // mstep)

                        # Generate the months for these steps
                        times = []
                        for month in range(mmin, mmax + 1, mstep):
                            times.append(datetime(year=ymin + month // 12,
                                                  month=month % 12, day=1))

                    # Convert back to MJD
                    values = Time(times, scale=self._converter.scale).mjd

                elif vrange > 1:  # greater than a day

                    self.set_params(steps=[1, 2, 5, 10])
                    values = super().tick_values(vmin, vmax)

                else:

                    # Determine ideal step
                    dv = vrange / 3 * 24 << u.hourangle

                    # And round to nearest sensible value
                    dv = select_step_hour(dv).to_value(u.hourangle) / 24

                    # Determine tick locations
                    imin = np.ceil(vmin / dv)
                    imax = np.floor(vmax / dv)
                    values = np.arange(imin, imax + 1, dtype=np.int64) * dv

            else:

                values = super().tick_values(vmin, vmax)

            # Get rid of values outside of the input interval
            values = values[(values >= vmin) & (values <= vmax)]

            return values

        def __call__(self):
            vmin, vmax = self.axis.get_view_interval()
            return self.tick_values(vmin, vmax)

    class AstropyTimeFormatter(ScalarFormatter):

        def __init__(self, converter, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._converter = converter
            self.set_useOffset(False)
            self.set_scientific(False)

        def __call__(self, value, pos=None):
            # Needed for Matplotlib <3.1
            if self._converter.format in STR_FORMATS:
                return self.format_ticks([value])[0]
            else:
                return super().__call__(value, pos=pos)

        def format_ticks(self, values):
            if len(values) == 0:
                return []
            if self._converter.format in YMDHMS_FORMATS:

                # Find spacing of labels in seconds - here we asssume the ticks
                # are evenly spaced, which they should be since we determined
                # them with the locator above.
                dv_sec = (values[1] - values[0]) * 3600 * 24
                if dv_sec < 1:
                    precision = int(np.ceil(-np.log10(dv_sec)))
                else:
                    precision = 0

                times = Time(values, format='mjd', scale=self._converter.scale, precision=precision)
                formatted = getattr(times, self._converter.format)
                if self._converter.simplify:
                    if self._converter.format in ('fits', 'iso', 'isot'):
                        separator = ' ' if self._converter.format == 'iso' else 'T'
                        # The regular expression is to check whether the time
                        # contains only zeros and separators.
                        if all([not re.match('[1-9]', x) for x in formatted]):
                            formatted = [x.split(separator)[0] for x in formatted]
                        else:
                            date_prev = None
                            for i in range(len(formatted)):
                                date, time = formatted[i].split(separator)
                                if date == date_prev:
                                    formatted[i] = time
                                date_prev = date
                    elif self._converter.format == 'yday':
                        if all([x.endswith(':001:00:00:00.000') for x in formatted]):
                            formatted = [x.split(':', 1)[0] for x in formatted]
                return formatted
            elif self._converter.format == 'byear_str':
                return Time(values, format='byear', scale=self._converter.scale).byear_str
            elif self._converter.format == 'jyear_str':
                return Time(values, format='jyear', scale=self._converter.scale).jyear_str
            else:
                return super().format_ticks(values)

    class MplTimeConverter(units.ConversionInterface):

        def __init__(self, scale=None, format=None, simplify=None):

            super().__init__()

            self.format = format
            self.scale = scale
            self.simplify = simplify

            # Keep track of original converter in case the context manager is
            # used in a nested way.
            self._original_converter = units.registry.get(Time)

            units.registry[Time] = self

        @property
        def format(self):
            return self._format

        @format.setter
        def format(self, value):
            if value in UNSUPPORTED_FORMATS:
                raise ValueError('time_support does not support format={0}'.format(value))
            self._format = value

        def __enter__(self):
            return self

        def __exit__(self, type, value, tb):
            if self._original_converter is None:
                del units.registry[Time]
            else:
                units.registry[Time] = self._original_converter

        def default_units(self, x, axis):
            if isinstance(x, tuple):
                x = x[0]
            if self.format is None:
                self.format = x.format
            if self.scale is None:
                self.scale = x.scale
            return 'astropy_time'

        def convert(self, value, unit, axis):
            """
            Convert a Time value to a scalar or array.
            """

            # For cases where Matplotlib doesn't implement the ConversionInterface
            if isinstance(value, (tuple, list)) and isinstance(value[0], Time):
                value = Time(value)

            # For Matplotlib < 2.2
            if not isinstance(value, Time):
                return value

            scaled = getattr(value, self.scale)
            if self.format in YMDHMS_FORMATS:
                return scaled.mjd
            elif self.format == 'byear_str':
                return scaled.byear
            elif self.format == 'jyear_str':
                return scaled.jyear
            else:
                return getattr(scaled, self.format)

        def axisinfo(self, unit, axis):
            """
            Return major and minor tick locators and formatters.
            """
            majloc = AstropyTimeLocator(self)
            majfmt = AstropyTimeFormatter(self)
            return units.AxisInfo(majfmt=majfmt,
                                  majloc=majloc,
                                  label='Time ({0})'.format(self.scale))

    return MplTimeConverter(scale=scale, format=format, simplify=simplify)



from collections import defaultdict

from palettable.colorbrewer.qualitative import Paired_5

from aas_timeseries.layers import Text, Markers, Line, VerticalLine, HorizontalLine

__all__ = ['auto_assign_colors']

ALWAYS_BLACK = (Text,)
BLACK_IF_ONLY_ONE = (Markers, Line, VerticalLine, HorizontalLine)


def auto_assign_colors(layers):

    colors = []

    layers_by_type = defaultdict(list)
    for layer in layers:
        layers_by_type[type(layer)].append(layer)

    offset = 0

    for layer in layers:
        if isinstance(layer, ALWAYS_BLACK):
            colors.append('#000000')
        elif isinstance(layer, BLACK_IF_ONLY_ONE) and len(layers_by_type[type(layer)]) == 1:
            colors.append('#000000')
        else:
            icolor = layers_by_type[type(layer)].index(layer)
            current = (offset + icolor) % 4
            colors.append(Paired_5.hex_colors[current])
            offset = current + 1

    return colors



import uuid
import weakref
from traitlets import HasTraits
from astropy import units as u
from aas_timeseries.traits import (Unicode, CFloat, PositiveCFloat, Opacity, Color,
                                   UnicodeChoice, DataTrait, ColumnTrait, AstropyTime,
                                   AstropyQuantity, Tooltip)

__all__ = ['BaseLayer', 'Markers', 'Line', 'Range', 'VerticalLine',
           'VerticalRange', 'HorizontalLine', 'HorizontalRange', 'Text',
           'time_to_vega', 'TimeDependentLayer']

DEFAULT_COLOR = '#000000'


def time_to_vega(time):
    """
    Convert an `~astropy.time.Time` object into a string compatible with Vega.
    """

    year, month, day, hour, minute, second = time.datetime.timetuple()[:6]

    # Note that Vega assumes months are zero-based.
    month -= 1

    return f'datetime({year}, {month}, {day}, {hour}, {minute}, {second})'


def generate_tooltip(tooltip_option, default_tooltip):

    if isinstance(tooltip_option, bool):
        if tooltip_option:
            return default_tooltip
        else:
            return None
    elif isinstance(tooltip_option, (tuple, list)):
        return {'signal': "{" + ', '.join(["'{0}': datum.{0}".format(col) for col in tooltip_option]) + "}"}
    elif isinstance(tooltip_option, dict):
        return {'signal': "{" + ', '.join(["'{0}': datum.{1}".format(val, key) for key, val in tooltip_option.items()]) + "}"}


class BaseLayer(HasTraits):
    """
    Base class for any layer object
    """

    n_uuids = 1

    label = Unicode(help='The label to use to designate the layers in the legend.')

    # Potential properties that could be implemented: toolTip

    def __init__(self, parent=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # NOTE: we use weakref to avoid circular references
        self.parent = weakref.ref(parent)
        self.uuids = [str(uuid.uuid4()) for i in range(self.n_uuids)]

    def remove(self):
        """
        Remove the layer from the visualization.
        """
        if self.parent is None or self.parent() is None:
            raise Exception(f"Layer '{self.label}' is no longer in a figure/view")
        else:
            self.parent().remove(self)
            self.parent = None

    def to_vega(self):
        """
        Convert the layer to its Vega representation.
        """

    def to_mpl(self, ax, yunit=None):
        """
        Add the layer to a Matplotlib `~matplotlib.axes.Axes` instance.
        """

    @property
    def _required_xdata(self):
        """
        Return a list of (data, column) tuples giving the data/columns required
        for the x-axis of the layer.
        """
        return []

    @property
    def _required_ydata(self):
        """
        Return a list of (data, column) tuples giving the data/columns required
        for the y-axis of the layer.
        """
        return []

    @property
    def _required_tooltipdata(self):
        """
        Return a list of (data, column) tuples giving the data/columns required
        for the tool tip of the layer.
        """
        return []


class TimeDependentLayer(BaseLayer):
    """
    A common class for all layers that depend on time
    """

    time_column = ColumnTrait(None, help='The column to use.')


MARKER_SHAPES = ['circle', 'square', 'cross', 'diamond', 'triangle-up',
                 'triangle-down', 'triangle-right', 'triangle-left']


class Markers(TimeDependentLayer):
    """
    A set of time series data points represented by markers.
    """

    n_uuids = 2

    data = DataTrait(help='The time series object containing the data.')
    column = ColumnTrait(None, help='The field in the time series containing the data.')
    error = ColumnTrait(None, help='The field in the time series '
                                   'containing the data uncertainties.')

    shape = UnicodeChoice('circle', help='The symbol shape.', choices=MARKER_SHAPES)

    size = PositiveCFloat(20, help='The area in pixels of the bounding box of the symbols.\n\n'
                                   'Note that this value sets the area of the symbol; the '
                                   'side lengths will increase with the square root of this '
                                   'value.')

    color = Color(None, help='The fill color of the symbols.')
    opacity = Opacity(1, help='The opacity of the fill color from 0 (transparent) to 1 (opaque).')

    edge_color = Color(None, help='The edge color of the symbol.')
    edge_opacity = Opacity(0.2, help='The opacity of the edge color from 0 (transparent) to 1 (opaque).')
    edge_width = PositiveCFloat(0, help='The thickness of the edge, in pixels.')

    error_width = PositiveCFloat(1, help='The width of the error bar, in pixels.')

    tooltip = Tooltip(True, help='Whether to show a tooltip (`False` or '
                                 '`True`). Can also be set to a list of data '
                                 'columns to show, or a dictionary mapping the '
                                 'display name to the column name.')

    def to_vega(self, yunit=None):

        default_tooltip = {'signal': "{{'{0}': datum.{0}, '{1}': datum.{1}}}".format(self.time_column, self.column)}

        # The main markers
        vega = [{'type': 'symbol',
                 'name': self.uuids[0],
                 'description': self.label,
                 'clip': True,
                 'from': {'data': self.data.uuid},
                 'encode': {'enter': {'x': {'scale': 'xscale', 'field': self.time_column},
                                      'y': {'scale': 'yscale', 'field': self.column},
                                      'shape': {'value': self.shape}},
                            'update': {'shape': {'value': self.shape},
                                       'size': {'value': self.size},
                                       'fill': {'value': self.color or DEFAULT_COLOR},
                                       'fillOpacity': {'value': self.opacity},
                                       'stroke': {'value': self.edge_color or DEFAULT_COLOR},
                                       'strokeOpacity': {'value': self.edge_opacity},
                                       'strokeWidth': {'value': self.edge_width}}}}]

        if self.tooltip:
            vega[0]['encode']['hover'] = {'size': {'value': self.size * 4},
                                          'tooltip': generate_tooltip(self.tooltip, default_tooltip)}

        # The error bars (if requested)
        if self.error:
            vega.append({'type': 'rect',
                         'name': self.uuids[1],
                         'description': self.label,
                         'clip': True,
                         'from': {'data': self.data.uuid},
                         'encode': {'enter': {'x': {'scale': 'xscale', 'field': self.time_column},
                                              'y': {'scale': 'yscale', 'signal': f"datum['{self.column}'] - datum['{self.error}']"},
                                              'y2': {'scale': 'yscale', 'signal': f"datum['{self.column}'] + datum['{self.error}']"}},
                                    'update': {'shape': {'value': self.shape},
                                               'width': {'value': self.error_width},
                                               'fill': {'value': self.color or DEFAULT_COLOR},
                                               'fillOpacity': {'value': self.opacity},
                                               'stroke': {'value': self.edge_color or DEFAULT_COLOR},
                                               'strokeOpacity': {'value': self.edge_opacity},
                                               'strokeWidth': {'value': self.edge_width}}}})

        return vega

    def to_mpl(self, ax, yunit=None):

        x = self.data.time_series[self.time_column]
        y = self.data.column_to_values(self.column, yunit)

        ax.scatter(x, y, s=self.size / 2,
                   color=self.color or DEFAULT_COLOR,
                   alpha=self.opacity)

        if self.error:
            yerr = self.data.column_to_values(self.error, yunit)
            ax.errorbar(x, y, yerr=yerr, fmt='none',
                        color=self.color or DEFAULT_COLOR,
                        alpha=self.opacity, linewidth=self.error_width)

    @property
    def _required_xdata(self):
        return [(self.data, self.time_column)]

    @property
    def _required_ydata(self):
        return [(self.data, self.column), (self.data, self.error)]

    @property
    def _required_tooltipdata(self):
        if isinstance(self.tooltip, bool):
            if self.tooltip:
                return self._required_xdata + self._required_ydata
            else:
                return []
        elif isinstance(self.tooltip, (tuple, list, dict)):
            return [(self.data, col) for col in self.tooltip]


class Line(TimeDependentLayer):
    """
    A set of time series data points connected by a line.
    """

    data = DataTrait(help='The time series object containing the data.')
    column = ColumnTrait(None, help='The field in the time series containing the data.')
    width = PositiveCFloat(1, help='The width of the line, in pixels.')

    color = Color(None, help='The color of the line.')
    opacity = Opacity(1, help='The opacity of the line from 0 (transparent) to 1 (opaque).')

    def to_vega(self, yunit=None):
        vega = {'type': 'line',
                'name': self.uuids[0],
                'description': self.label,
                'clip': True,
                'from': {'data': self.data.uuid},
                'encode': {'enter': {'x': {'scale': 'xscale', 'field': self.time_column},
                                     'y': {'scale': 'yscale', 'field': self.column},
                                     'stroke': {'value': self.color or DEFAULT_COLOR},
                                     'strokeOpacity': {'value': self.opacity},
                                     'strokeWidth': {'value': self.width}}}}
        return [vega]

    def to_mpl(self, ax, yunit=None):

        x = self.data.time_series[self.time_column]
        y = self.data.column_to_values(self.column, yunit)

        ax.plot(x, y, '-',
                linewidth=self.width,
                color=self.color or DEFAULT_COLOR,
                alpha=self.opacity)

    @property
    def _required_xdata(self):
        return [(self.data, self.time_column)]

    @property
    def _required_ydata(self):
        return [(self.data, self.column)]


class Range(TimeDependentLayer):
    """
    An interval defined by lower and upper values as a function of time.
    """

    data = DataTrait(help='The time series object containing the data.')
    column_lower = ColumnTrait(None, help='The field in the time series containing the lower value of the data range.')
    column_upper = ColumnTrait(None, help='The field in the time series containing the upper value of the data range.')

    color = Color(None, help='The fill color of the range.')
    opacity = Opacity(0.2, help='The opacity of the fill color from 0 (transparent) to 1 (opaque).')

    edge_color = Color(None, help='The edge color of the range.')
    edge_opacity = Opacity(0.2, help='The opacity of the edge color from 0 (transparent) to 1 (opaque).')
    edge_width = PositiveCFloat(0, help='The thickness of the edge, in pixels.')

    # Potential properties that could be implemented: strokeCap, strokeDash

    def to_vega(self, yunit=None):
        vega = {'type': 'area',
                'name': self.uuids[0],
                'description': self.label,
                'clip': True,
                'from': {'data': self.data.uuid},
                'encode': {'enter': {'x': {'scale': 'xscale', 'field': self.time_column},
                                     'y': {'scale': 'yscale', 'field': self.column_lower},
                                     'y2': {'scale': 'yscale', 'field': self.column_upper},
                                     'fill': {'value': self.color or DEFAULT_COLOR},
                                     'fillOpacity': {'value': self.opacity},
                                     'stroke': {'value': self.edge_color or DEFAULT_COLOR},
                                     'strokeOpacity': {'value': self.edge_opacity},
                                     'strokeWidth': {'value': self.edge_width}}}}

        return [vega]

    def to_mpl(self, ax, yunit=None):

        x = self.data.time_series[self.time_column]
        y1 = self.data.column_to_values(self.column_lower, yunit)
        y2 = self.data.column_to_values(self.column_upper, yunit)

        ax.fill_between(x, y1, y2,
                        color=self.color or DEFAULT_COLOR,
                        alpha=self.opacity)

    @property
    def _required_xdata(self):
        return [(self.data, self.time_column)]

    @property
    def _required_ydata(self):
        return [(self.data, self.column_lower), (self.data, self.column_upper)]


class VerticalLine(BaseLayer):
    """
    A vertical line at a specific time.
    """

    time = AstropyTime(help='The date/time at which the vertical line is shown.')
    width = PositiveCFloat(1, help='The width of the line, in pixels.')

    color = Color(None, help='The color of the line.')
    opacity = Opacity(1, help='The opacity of the line from 0 (transparent) to 1 (opaque).')

    # Potential properties that could be implemented: strokeCap, strokeDash

    def to_vega(self, yunit=None):

        vega = {'type': 'rule',
                'name': self.uuids[0],
                'description': self.label,
                'clip': True,
                'encode': {'enter': {'x': {'scale': 'xscale', 'signal': time_to_vega(self.time)},
                                     'y': {'value': 0},
                                     'y2': {'field': {'group': 'height'}},
                                     'strokeWidth': {'value': self.width},
                                     'stroke': {'value': self.color or DEFAULT_COLOR},
                                     'strokeOpacity': {'value': self.opacity}}}}
        return [vega]

    def to_mpl(self, ax, yunit=None):
        ax.axvline(self.time,
                   linewidth=self.width,
                   color=self.color or DEFAULT_COLOR,
                   alpha=self.opacity)


class VerticalRange(BaseLayer):
    """
    A continuous range specified by a lower and upper time.
    """

    time_lower = AstropyTime(help='The date/time at which the range starts.')
    time_upper = AstropyTime(help='The date/time at which the range ends.')

    color = Color(None, help='The fill color of the range.')
    opacity = Opacity(0.2, help='The opacity of the fill color from 0 (transparent) to 1 (opaque).')

    edge_color = Color(None, help='The edge color of the range.')
    edge_opacity = Opacity(0.2, help='The opacity of the edge color from 0 (transparent) to 1 (opaque).')
    edge_width = PositiveCFloat(0, help='The thickness of the edge, in pixels.')

    # Potential properties that could be implemented: strokeCap, strokeDash

    def to_vega(self, yunit=None):

        vega = {'type': 'rect',
                'name': self.uuids[0],
                'description': self.label,
                'clip': True,
                'encode': {'enter': {'x': {'scale': 'xscale', 'signal': time_to_vega(self.time_lower)},
                                     'x2': {'scale': 'xscale', 'signal': time_to_vega(self.time_upper)},
                                     'y': {'value': 0},
                                     'y2': {'field': {'group': 'height'}},
                                     'fill': {'value': self.color or DEFAULT_COLOR},
                                     'fillOpacity': {'value': self.opacity},
                                     'stroke': {'value': self.edge_color or DEFAULT_COLOR},
                                     'strokeOpacity': {'value': self.edge_opacity},
                                     'strokeWidth': {'value': self.edge_width}}}}

        return [vega]

    def to_mpl(self, ax, yunit=None):
        ax.fill_betweenx([-1e30, 1e30], self.time_lower, self.time_upper,
                         color=self.color or DEFAULT_COLOR,
                         alpha=self.opacity)


class HorizontalLine(BaseLayer):
    """
    A horizontal line at a specific y value.
    """

    # TODO: validate value and allow it to be a quantity
    value = AstropyQuantity(help='The y value at which the horizontal line is shown.')
    width = PositiveCFloat(1, help='The width of the line, in pixels.')

    color = Color(None, help='The color of the line.')
    opacity = Opacity(1, help='The opacity of the line from 0 (transparent) to 1 (opaque).')

    # Potential properties that could be implemented: strokeCap, strokeDash

    def to_vega(self, yunit=None):

        if yunit is None:
            yunit = u.one

        value = self.value.to_value(yunit)

        vega = {'type': 'rule',
                'name': self.uuids[0],
                'description': self.label,
                'clip': True,
                'encode': {'enter': {'x': {'value': 0},
                                     'x2': {'field': {'group': 'width'}},
                                     'y': {'scale': 'yscale', 'value': value},
                                     'strokeWidth': {'value': self.width},
                                     'stroke': {'value': self.color or DEFAULT_COLOR},
                                     'strokeOpacity': {'value': self.opacity}}}}
        return [vega]

    def to_mpl(self, ax, yunit=None):
        if yunit is None:
            yunit = u.one
        ax.axhline(self.value.to_value(yunit),
                   linewidth=self.width,
                   color=self.color or DEFAULT_COLOR,
                   alpha=self.opacity)


class HorizontalRange(BaseLayer):
    """
    A continuous range specified by a lower and upper value.
    """

    value_lower = AstropyQuantity(help='The value at which the range starts.')
    value_upper = AstropyQuantity(help='The value at which the range ends.')

    color = Color(None, help='The fill color of the range.')
    opacity = Opacity(0.2, help='The opacity of the fill color from 0 (transparent) to 1 (opaque).')

    edge_color = Color(None, help='The edge color of the range.')
    edge_opacity = Opacity(0.2, help='The opacity of the edge color from 0 (transparent) to 1 (opaque).')
    edge_width = PositiveCFloat(0, help='The thickness of the edge, in pixels.')

    # Potential properties that could be implemented: strokeCap, strokeDash

    def to_vega(self, yunit=None):

        if yunit is None:
            yunit = u.one

        value_lower = self.value_lower.to_value(yunit)
        value_upper = self.value_upper.to_value(yunit)

        vega = {'type': 'rect',
                'name': self.uuids[0],
                'description': self.label,
                'clip': True,
                'encode': {'enter': {'x': {'value': 0},
                                     'x2': {'field': {'group': 'width'}},
                                     'y': {'scale': 'yscale', 'value': value_lower},
                                     'y2': {'scale': 'yscale', 'value': value_upper},
                                     'fill': {'value': self.color or DEFAULT_COLOR},
                                     'fillOpacity': {'value': self.opacity},
                                     'stroke': {'value': self.edge_color or DEFAULT_COLOR},
                                     'strokeOpacity': {'value': self.edge_opacity},
                                     'strokeWidth': {'value': self.edge_width}}}}
        return [vega]

    def to_mpl(self, ax, yunit=None):

        value_lower = self.value_lower.to_value(yunit)
        value_upper = self.value_upper.to_value(yunit)

        ax.fill_between([-1e30, 1e30], value_lower, value_upper,
                        color=self.color or DEFAULT_COLOR,
                        alpha=self.opacity)


class Text(BaseLayer):
    """
    A text label.
    """

    text = Unicode(help='The text label to show.')
    time = AstropyTime(help='The date/time at which the text is shown.')
    value = AstropyQuantity(help='The y value at which the text is shown.')
    weight = UnicodeChoice('normal', help='The weight of the text.',
                           choices=['normal', 'bold'])
    baseline = UnicodeChoice('alphabetic', help='The vertical text baseline.',
                             choices=['alphabetic', 'top', 'middle', 'bottom'])
    align = UnicodeChoice('left', help='The horizontal text alignment.',
                          choices=['left', 'center', 'right'])
    angle = CFloat(0, help='The rotation angle of the text in degrees (default 0).')

    color = Color(None, help='The color of the text.')
    opacity = Opacity(1, help='The opacity of the text from 0 (transparent) to 1 (opaque).')

    def to_vega(self, yunit=None):

        if yunit is None:
            yunit = u.one

        value = self.value.to_value(yunit)

        vega = {'type': 'text',
                'name': self.uuids[0],
                'description': self.label,
                'clip': True,
                'encode': {'enter': {'x': {'scale': 'xscale', 'signal': time_to_vega(self.time)},
                                     'y': {'scale': 'yscale', 'value': value},
                                     'text': {'value': self.text},
                                     'fill': {'value': self.color or DEFAULT_COLOR},
                                     'fillOpacity': {'value': self.opacity},
                                     'fontWeigth': {'value': self.weight},
                                     'baseline': {'value': self.baseline},
                                     'align': {'value': self.align},
                                     'angle': {'value': self.angle}}}}
        return [vega]

    def to_mpl(self, ax, yunit=None):

        if yunit is None:
            yunit = u.one

        x = self.time
        value = self.value.to_value(yunit)

        ax.text(x, value, self.text,
                color=self.color or DEFAULT_COLOR,
                alpha=self.opacity)



import uuid

from astropy.units import Quantity, UnitsError

__all__ = ['Data']


class Data:

    def __init__(self, time_series):
        self.time_series = time_series
        self.uuid = str(uuid.uuid4())
        self.time_column = 'time'

    def column_to_values(self, colname, unit):

        # First make sure the column is a quantity
        quantity = Quantity(self.time_series[colname], copy=False)

        if quantity.unit.is_equivalent(unit):
            return quantity.to_value(unit)
        else:
            raise UnitsError(f"Cannot convert the units '{quantity.unit}' of "
                             f"column '{colname}' to the required units of "
                             f"'{unit}'")

    def unit(self, colname):
        return Quantity(self.time_series[colname], copy=False).unit



from qtpy.QtWebEngineWidgets import QWebEngineView, QWebEnginePage, WEBENGINE
from qtpy import QtGui, QtWidgets, QtCore

__all__ = ['get_qt_web_widget']


class TimeSeriesWebEnginePage(QWebEnginePage):
    """
    Subclass of QWebEnginePage that can check when WWT is ready for
    commands.
    """

    wwt_ready = QtCore.Signal()

    def __init__(self, parent=None):
        super(TimeSeriesWebEnginePage, self).__init__(parent=parent)
        if not WEBENGINE:
            self._frame = self.mainFrame()

    if WEBENGINE:

        def javaScriptConsoleMessage(self, level=None, message=None,
                                     line_number=None, source_id=None):
            print(f'{message} (level={level}, line_number={line_number}, '
                  f'source_id={source_id})')

        def _process_js_response(self, result):
            self._js_response_received = True
            self._js_response = result

        def runJavaScript(self, code, asynchronous=True):
            app = QtWidgets.QApplication.instance()
            if asynchronous:
                super(TimeSeriesWebEnginePage, self).runJavaScript(code)
            else:
                self._js_response_received = False
                self._js_response = None
                super(TimeSeriesWebEnginePage, self).runJavaScript(code, self._process_js_response)
                while not self._js_response_received:
                    app.processEvents()
                return self._js_response

    else:

        def javaScriptConsoleMessage(self, message=None, line_number=None,
                                     source_id=None):
            print(f'{message} (line_number={line_number}, '
                  f'source_id={source_id})')


class TimeSeriesWebEngineView(QWebEngineView):

    def save_to_file(self, filename):
        image = QtGui.QImage(self.size(), QtGui.QImage.Format_RGB32)
        painter = QtGui.QPainter(image)
        self.render(painter)
        image.save(filename)
        painter.end()


def get_qt_web_widget(url):
    web = TimeSeriesWebEngineView()
    page = TimeSeriesWebEnginePage()
    page.setView(web)
    web.setPage(page)
    web.setUrl(QtCore.QUrl(url))
    return web, page



import os
import pytest
from traitlets import TraitError

from astropy import units as u
from astropy.timeseries import TimeSeries

from aas_timeseries.visualization import InteractiveTimeSeriesFigure
from aas_timeseries.screenshot import interactive_screenshot
from aas_timeseries.tests.helpers import compare_to_reference_json, DATA


class TestFigure:

    def setup_method(self):

        self.ts = TimeSeries(time_start='2016-03-22T12:30:31', time_delta=3 * u.s, n_samples=5)
        self.ts['flux'] = [1, 2, 3, 4, 5]
        self.ts['error'] = [1, 2, 3, 4, 5]

    def test_basic(self, tmpdir, deterministic_uuid, image_tests):

        # A test that uses all the types of layers. This includes an optional
        # pixel-by-pixel image test that is enabled when using the --image-tests
        # command-line argument.

        figure = InteractiveTimeSeriesFigure()
        figure.add_markers(time_series=self.ts, column='flux', label='Markers')
        figure.add_line(time_series=self.ts, column='flux', label='Line')
        figure.add_markers(time_series=self.ts, column='flux', error='error', label='Markers with Errors')
        figure.add_vertical_line(self.ts.time[3], label='Vertical Line')
        figure.add_vertical_range(self.ts.time[0], self.ts.time[-1], label='Vertical Range')
        figure.add_horizontal_line(3, label='Horizontal Line')
        figure.add_horizontal_range(5, 6, label='Horizontal Range')
        figure.add_range(time_series=self.ts, column_lower='flux', column_upper='error', label='Range')
        figure.add_text(time=self.ts.time[2], value=float(self.ts['flux'][0]), text='My Label', label='Range')
        figure.ylabel = 'Intensity'

        json_file = tmpdir.join('figure.json').strpath
        plot_prefix = tmpdir.join('figure').strpath
        figure.save_vega_json(json_file)
        if image_tests:
            interactive_screenshot(json_file, plot_prefix)
            figure.save_static(plot_prefix + '_mpl', format='png')

        compare_to_reference_json(tmpdir, 'basic', image_tests=image_tests)

    def test_save_options_embed(self, tmpdir, deterministic_uuid):

        # Make sure that the data is correctly embedded inside the JSON file

        figure = InteractiveTimeSeriesFigure()
        figure.add_markers(time_series=self.ts, column='flux', label='Markers')
        figure.save_vega_json(tmpdir.join('figure.json').strpath, embed_data=True)
        compare_to_reference_json(tmpdir, 'save_options_embed')

    def test_save_options_no_minimize(self, tmpdir, deterministic_uuid):

        # Make sure that the data contains all columns, even those not needed

        figure = InteractiveTimeSeriesFigure()
        figure.add_markers(time_series=self.ts, column='flux', label='Markers')

        figure.save_vega_json(tmpdir.join('figure.json').strpath, minimize_data=False)
        compare_to_reference_json(tmpdir, 'save_options_no_minimize')

    def test_save_options_export_bundle(self, tmpdir):

        # Test saving the figure to a zip bundle

        figure = InteractiveTimeSeriesFigure()
        figure.add_markers(time_series=self.ts, column='flux', label='Markers')

        figure.export_interactive_bundle(tmpdir.join('figure.zip').strpath)

    def test_column_validation(self):

        # Test the validation provied by ColumnTrait

        figure = InteractiveTimeSeriesFigure()
        with pytest.raises(TraitError) as exc:
            figure.add_markers(time_series=self.ts, column='flux2', label='Markers')
        assert exc.value.args[0] == 'flux2 is not a valid column name'

    def test_limits(self, tmpdir):

        # Test the validation when setting limits

        filename = tmpdir.join('figure.json').strpath

        figure = InteractiveTimeSeriesFigure()
        figure.add_markers(time_series=self.ts, column='flux', label='Markers')
        figure.xlim = self.ts.time[0], self.ts.time[-1]
        figure.ylim = 0, 10
        figure.save_vega_json(filename)

        with pytest.raises(TypeError) as exc:
            figure.xlim = 0, 1
        assert exc.value.args[0] == 'xlim should be a typle of two Time instances'

        with pytest.raises(ValueError) as exc:
            figure.xlim = self.ts.time[0], self.ts.time[-1], self.ts.time[-4]
        assert exc.value.args[0] == 'xlim should be a tuple of two elements'

    def test_views(self, tmpdir):

        # Make sure that views contain the right layers

        filename = tmpdir.join('figure.json').strpath

        figure = InteractiveTimeSeriesFigure()

        markers = figure.add_markers(time_series=self.ts, column='flux', label='Markers')
        line = figure.add_line(time_series=self.ts, column='flux', label='Line')

        # By default views inherit all layers from base figure
        view1 = figure.add_view('Test1')
        assert figure.layers == [markers, line]
        assert view1.layers == [markers, line]

        # And we can add view-specific layers to them
        vertical = view1.add_vertical_line(self.ts.time[3], label='Vertical Line')
        assert figure.layers == [markers, line]
        assert view1.layers == [markers, line, vertical]

        # But adding layers to the figure after the view is created doesnt' cause
        # them to get added to the view
        horizontal = figure.add_horizontal_line(3., label='Horizontal Line')
        assert figure.layers == [markers, line, horizontal]
        assert view1.layers == [markers, line, vertical]

        # We can use include to specify which initial layers to include
        view2 = figure.add_view('Test1', include=[markers])
        assert view2.layers == [markers]

        # and exclude to, well, exclude layers
        view3 = figure.add_view('Test1', exclude=[markers])
        assert view3.layers == [line, horizontal]

        # We also provide an 'empty' shortcut that means include=[]
        view4 = figure.add_view('Test1', empty=True)
        assert view4.layers == []

        # We tell the user if the include or exclude list contain invalid values
        with pytest.raises(ValueError) as exc:
            figure.add_view('Test1', exclude=[vertical])
        assert 'does not exist in base figure' in exc.value.args[0]
        with pytest.raises(ValueError) as exc:
            figure.add_view('Test1', include=[vertical])
        assert 'does not exist in base figure' in exc.value.args[0]

        figure.save_vega_json(filename)

    def test_remove(self):

        # Make sure that removing layers works correctly

        figure = InteractiveTimeSeriesFigure()
        assert len(figure.layers) == 0
        markers = figure.add_markers(time_series=self.ts, column='flux', label='Markers')
        assert len(figure.layers) == 1
        line = figure.add_line(time_series=self.ts, column='flux', label='Line')
        assert len(figure.layers) == 2

        view = figure.add_view('Test view')
        assert len(view.layers) == 2

        range = view.add_vertical_range(self.ts.time[0], self.ts.time[-1], label='Vertical Range')

        assert len(figure.layers) == 2
        assert len(view.layers) == 3

        # Removing using the .remove() method on a layer removes it from the
        # figure and all views where it is.
        line.remove()
        assert len(figure.layers) == 1
        assert len(view.layers) == 2

        # Removing from the .remove() method on a view removes it just from the view
        view.remove(markers)
        assert len(figure.layers) == 1
        assert len(view.layers) == 1

        # Check removing a view-specific layer
        range.remove()
        assert len(figure.layers) == 1
        assert len(view.layers) == 0

        # Remove last layer from figure
        markers.remove()

        with pytest.raises(Exception) as exc:
            markers.remove()
        assert exc.value.args[0] == "Layer 'Markers' is no longer in a figure/view"

        with pytest.raises(Exception) as exc:
            figure.remove(markers)
        assert exc.value.args[0] == "Layer 'Markers' is not in figure"

        with pytest.raises(Exception) as exc:
            view.remove(markers)
        assert exc.value.args[0] == "Layer 'Markers' is not in view"


class TestUnit:

    # Make sure that things work as expected when units are present, and that
    # the correct errors are raised when mixing data with and without units.

    def setup_method(self, method):

        self.ts = TimeSeries(time_start='2016-03-22T12:30:31', time_delta=3 * u.s, n_samples=5)
        self.ts['flux'] = [1, 2, 3, 4, 5]
        self.ts['error'] = [1, 2, 3, 4, 5]
        self.ts['flux_with_unit'] = [1, 2, 3, 4, 5] * u.Jy
        self.ts['error_with_unit'] = [1, 2, 3, 4, 5] * u.mJy

        self.figure = InteractiveTimeSeriesFigure()

    def test_basic(self, tmpdir):
        # Make sure things work for a simple example with just one layer
        self.figure.add_markers(time_series=self.ts, column='flux_with_unit', label='Markers')
        self.figure.save_vega_json(tmpdir.join('figure.json').strpath)

    def test_all_markers(self, tmpdir):
        # A test with all the layer types and convertible units
        self.figure.add_markers(time_series=self.ts, column='flux_with_unit', label='Markers')
        self.figure.add_line(time_series=self.ts, column='flux_with_unit', label='Line')
        self.figure.add_markers(time_series=self.ts, column='flux_with_unit', error='error_with_unit', label='Markers with Errors')
        self.figure.add_vertical_line(self.ts.time[3], label='Vertical Line')
        self.figure.add_vertical_range(self.ts.time[0], self.ts.time[-1], label='Vertical Range')
        self.figure.add_horizontal_line(3 * u.MJy, label='Horizontal Line')
        self.figure.add_horizontal_range(5000 * u.mJy, 6 * u.Jy, label='Horizontal Range')
        self.figure.add_range(time_series=self.ts, column_lower='flux_with_unit', column_upper='error_with_unit', label='Range')
        self.figure.add_text(time=self.ts.time[2], value=self.ts['flux_with_unit'][0], text='My Label', label='Range')
        self.figure.save_vega_json(tmpdir.join('figure.json').strpath)

    def test_compatible_limits(self, tmpdir):
        # Set the limits using compatible units
        self.figure.add_markers(time_series=self.ts, column='flux_with_unit', label='Markers')
        self.figure.ylim = (1e-6 * u.MJy, 6000 * u.mJy)
        view = self.figure.add_view('my view')
        view.ylim = (2e-6 * u.MJy, 4000 * u.mJy)
        self.figure.save_vega_json(tmpdir.join('figure.json').strpath)

    def test_incompatible_limits(self, tmpdir):
        # Set the limits using incompatible units
        self.figure.add_markers(time_series=self.ts, column='flux_with_unit', label='Markers')
        self.figure.ylim = 1, 6
        with pytest.raises(u.UnitsError) as exc:
            self.figure.save_vega_json(tmpdir.join('figure.json').strpath)
        assert exc.value.args[0] == 'Limits for y axis are dimensionless but expected units of Jy'

    def test_incompatible_limits_view(self, tmpdir):
        # Set the limits using incompatible units (with a view)
        self.figure.add_markers(time_series=self.ts, column='flux_with_unit', label='Markers')
        view = self.figure.add_view('my view')
        view.ylim = 1, 6
        with pytest.raises(u.UnitsError) as exc:
            self.figure.save_vega_json(tmpdir.join('figure.json').strpath)
        assert exc.value.args[0] == "Limits for y axis are dimensionless but expected units of Jy"

    def test_two_marker_layer_incompatible(self, tmpdir):
        self.figure.add_markers(time_series=self.ts, column='flux_with_unit', label='Markers')
        self.figure.add_markers(time_series=self.ts, column='flux', label='Markers')
        with pytest.raises(u.UnitsError) as exc:
            self.figure.save_vega_json(tmpdir.join('figure.json').strpath)
        assert exc.value.args[0] == "Cannot convert the units '' of column 'flux' to the required units of 'Jy'"

    def test_incompatible_overlays(self, tmpdir):
        self.figure.add_markers(time_series=self.ts, column='flux_with_unit', label='Markers')
        self.figure.add_markers(time_series=self.ts, column='flux', label='Markers')
        with pytest.raises(u.UnitsError) as exc:
            self.figure.save_vega_json(tmpdir.join('figure.json').strpath)
        assert exc.value.args[0] == "Cannot convert the units '' of column 'flux' to the required units of 'Jy'"

    def test_custom_yunit(self, tmpdir):
        self.figure.add_markers(time_series=self.ts, column='flux_with_unit', label='Markers')
        self.figure.yunit = u.MJy
        self.figure.save_vega_json(tmpdir.join('figure.json').strpath)
        self.figure.yunit = 'mJy'  # with a string
        self.figure.save_vega_json(tmpdir.join('figure.json').strpath)

    def test_custom_incompatible_yunit(self, tmpdir):
        self.figure.add_markers(time_series=self.ts, column='flux_with_unit', label='Markers')
        self.figure.yunit = u.m
        with pytest.raises(u.UnitsError) as exc:
            self.figure.save_vega_json(tmpdir.join('figure.json').strpath)
        assert exc.value.args[0] == "Cannot convert the units 'Jy' of column 'flux_with_unit' to the required units of 'm'"

    def test_empty_base_figure(self, tmpdir):
        # Check that things work fine if the only layer is in a view
        view = self.figure.add_view('my view')
        view.add_markers(time_series=self.ts, column='flux_with_unit', label='Markers')
        self.figure.save_vega_json(tmpdir.join('figure.json').strpath)

    def test_no_data_layers(self, tmpdir):
        # In the case where there are no layers that depend on data, the default
        # assumed unit is u.one so adding a line with a unit will raise an error.
        # In future we might want to consider making this work.
        self.figure.add_horizontal_line(3 * u.mJy, label='Line')
        with pytest.raises(u.UnitsError) as exc:
            self.figure.save_vega_json(tmpdir.join('figure.json').strpath)
        assert exc.value.args[0] == "'mJy' (spectral flux density) and '' (dimensionless) are not convertible"

    def test_tooltip(self, tmpdir, deterministic_uuid):

        # Check that if a custom tooltip is specified, which uses columns other
        # than the ones used for the plotting, these other columns are included
        # in the output data file.

        self.ts['other1'] = [1, 2, 3, 4, 5]
        self.ts['other2'] = [1, 2, 3, 4, 5]
        self.ts['other3'] = [1, 2, 3, 4, 5]
        self.ts['other4'] = [1, 2, 3, 4, 5]

        figure = InteractiveTimeSeriesFigure()
        figure.add_markers(time_series=self.ts, column='flux', label='Markers',
                           tooltip=['time', 'flux', 'other1'])
        figure.add_markers(time_series=self.ts, column='flux', label='Markers',
                           tooltip={'other3': 'Other'})

        json_file = tmpdir.join('figure.json').strpath

        figure.save_vega_json(json_file)

        ts = TimeSeries.read(tmpdir.join('data_daea58ba-4c73-4942-8d87-78e7d340bbcd.csv').strpath,
                             format='ascii.basic', delimiter=',')

        assert ts.colnames == ['time', 'flux', 'other1', 'other3']


class TestTimeAxes:

    # Tests related to different time axes

    def setup_method(self):
        self.ts = TimeSeries.read(os.path.join(DATA, 'phase.csv'), format='ascii.csv')
        self.ts['error'] = 0.02  # reduce for now
        self.ts['relative_s'] = self.ts['relative'] * u.s

    def test_mixed_time_axes(self, tmpdir, deterministic_uuid, image_tests):

        # An extensive test that includes different types of time axes in
        # different views.

        fig = InteractiveTimeSeriesFigure(title='By time of observation')
        fig.ylabel = 'Radial Velocity (km/s)'

        # Absolute time
        fig.add_markers(time_series=self.ts, column='flux', color='orange', error='error', size=50)
        fig.add_line(time_series=self.ts, column='flux', color='black')

        # Phase
        view1 = fig.add_view(title='By phase', empty=True, time_mode='phase')
        view1.add_markers(time_series=self.ts, time_column='phase', column='flux', color='orange', error='error', size=50)

        # Relative time
        view2 = fig.add_view(title='By relative time', empty=True, time_mode='relative')
        view2.add_markers(time_series=self.ts, time_column='relative_s', column='flux', color='orange', error='error', size=50)

        # Absolute time (with julian date formatting)
        view3 = fig.add_view(title='Julian Date formatting')
        view3.time_format = 'jd'

        # Phase (with degrees formatting)
        view4 = fig.add_view(title='By phase', empty=True, time_mode='phase')
        view4.add_markers(time_series=self.ts, time_column='phase', column='flux', color='orange', error='error', size=50)
        view4.time_format = 'degrees'

        # Phase (with radians formatting)
        view5 = fig.add_view(title='By phase', empty=True, time_mode='phase')
        view5.add_markers(time_series=self.ts, time_column='phase', column='flux', color='orange', error='error', size=50)
        view5.time_format = 'radians'

        json_file = tmpdir.join('figure.json').strpath
        plot_prefix = tmpdir.join('figure').strpath
        fig.save_vega_json(json_file)
        if image_tests:
            interactive_screenshot(json_file, plot_prefix)
            fig.save_static(plot_prefix + '_mpl', format='png')

        compare_to_reference_json(tmpdir, 'mixed_time_axes', image_tests=image_tests)



# The main function in this module takes a JSON file and renders it to
# a PNG file. This uses Qt to launch a WebEngine widget, and serves the
# required files using tornado, then saves the screenshot of the contents
# of the widget with Qt.

import os
import json
import shutil
import time
import tempfile

from qtpy import QtWidgets

from aas_timeseries.screenshot.data_server import get_data_server
from aas_timeseries.screenshot.qt_web_widget import get_qt_web_widget

__all__ = ['interactive_screenshot']

ROOT = os.path.dirname(__file__)
TIMEOUT = 60  # seconds
WAIT_TIME = 0.1  # seconds

SET_VIEW_CODE = """
var view_ready = false;

function on_view_ready() {{
    view_ready = true;
}}

figure.setView({0}, {{'callback': on_view_ready}});
"""


def wait_for_true(app, page, var):
    """
    Wait until variable ``var`` is ``true`` in Javascript.
    """
    start = time.time()
    while time.time() - start < TIMEOUT:
        app.processEvents()
        ready = page.runJavaScript('(typeof({0}) != "undefined") && {0} == 1;'.format(var), asynchronous=False)
        if ready:
            break
    else:
        raise ValueError("Timed out while waiting for {0}==true".format(var))

    # Wait a little longer just to be sure
    start = time.time()
    while time.time() - start < WAIT_TIME:
        app.processEvents()


def interactive_screenshot(json_filename, prefix):
    """
    Given a JSON file, save the figure to one or more PNG files. If multiple
    views are present, each view will result in a separate PNG file.
    """

    tmpdir = tempfile.mkdtemp()
    tmp_html = os.path.join(tmpdir, 'page.html')
    tmp_json = os.path.join(tmpdir, 'figure.json')

    shutil.copy(os.path.join(ROOT, 'template.html'), tmp_html)
    shutil.copy(json_filename, tmp_json)

    server = get_data_server()
    url = server.serve_file(tmp_html)
    server.serve_file(tmp_json)

    # Check if we need to serve any csv files
    with open(json_filename) as f:
        figure = json.load(f)
    for data in figure['data']:
        if 'url' in data:
            server.serve_file(os.path.join(os.path.dirname(json_filename), data['url']))

    app = QtWidgets.QApplication.instance()
    if app is None:
        app = QtWidgets.QApplication([''])

    web, page = get_qt_web_widget(url)
    web.resize(figure['width'], figure['height'])
    web.show()

    # Wait for figure to be ready

    wait_for_true(app, page, 'figure_ready')

    web.save_to_file(prefix + '.png')

    # Find the views that are present in the figure
    views = page.runJavaScript('figure.getViews();', asynchronous=False)

    if len(views) > 1:
        for view_index in range(1, len(views)):

            page.runJavaScript(SET_VIEW_CODE.format(view_index), asynchronous=False)

            wait_for_true(app, page, 'view_ready')

            web.save_to_file(prefix + '_view{0}.png'.format(view_index))

    web.close()
    app.processEvents()

    # We need to do this to force garbage collection and avoid a
    # segmentation fault.
    page = web = None  # noqa




import os
import time
import socket
import logging
import asyncio
from hashlib import md5
from threading import Thread

__all__ = ['get_data_server']


def get_data_server(verbose=False):
    """
    This starts up a tornado server and returns a handle to a DataServer
    object which can be used to register files to serve.
    """

    from tornado.ioloop import IOLoop
    from tornado.web import RequestHandler, Application
    from tornado.routing import PathMatches

    class WebServer(Application):

        host = None
        port = None

        def run(self, host=None, port=8886):
            self.host = host
            self.port = port
            try:
                self.listen(port)
                IOLoop.instance().start()
            finally:
                self.host = None
                self.port = None

    class DataServer(object):

        def start(self, app):
            self._files = {}
            self._thread = Thread(target=self.start_app)
            self._thread.daemon = True
            self._thread.start()
            self._app = app
            while self._app.host is None and self._app.port is None:
                time.sleep(0.1)

        @property
        def port(self):
            return self._app.port

        @property
        def host(self):
            return self._app.host

        def start_app(self):

            asyncio.set_event_loop(asyncio.new_event_loop())

            host = socket.gethostbyname('localhost')

            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.bind(('localhost', 0))
            port = sock.getsockname()[1]
            sock.close()

            access_log = logging.getLogger("tornado.access")
            access_log.setLevel('ERROR')

            self._app.run(host=host, port=port)

        def serve_file(self, filename, real_name=True, extension=''):
            with open(filename, 'rb') as f:
                content = f.read()
            if real_name:
                hash = os.path.basename(filename)
            else:
                hash = md5(content).hexdigest() + extension
            self._files[hash] = os.path.abspath(filename)
            return 'http://' + self.host + ':' + str(self.port) + '/data/' + hash

        def get_file_contents(self, hash):
            with open(self._files[hash], 'rb') as f:
                return f.read()

    ds = DataServer()

    class DataHandler(RequestHandler):
        async def get(self, hash):
            self.write(ds.get_file_contents(hash))

    app = WebServer([(PathMatches(r"/data/(?P<hash>\S+)"), DataHandler)])

    ds.start(app)

    return ds



from fractions import Fraction
import numpy as np
from matplotlib.ticker import MaxNLocator, StrMethodFormatter, Formatter

__all__ = ['PhaseAsDegreesLocator', 'PhaseAsDegreesFormatter',
           'PhaseAsRadiansLocator', 'PhaseAsRadiansFormatter']


class PhaseAsDegreesLocator(MaxNLocator):
    """
    Matplotlib locator for phases in the range [0:1] as degrees.
    """

    def __init__(self, *args, **kwargs):
        kwargs['nbins'] = 6
        super().__init__(*args, **kwargs)

    def tick_values(self, vmin, vmax):
        values = super().tick_values(vmin * 360, vmax * 360) / 360.
        return values[(values >= vmin) & (values <= vmax)]

    def __call__(self):
        vmin, vmax = self.axis.get_view_interval()
        return self.tick_values(vmin, vmax)


class PhaseAsDegreesFormatter(StrMethodFormatter):
    """
    Matplotlib formatter for phases in the range [0:1] as degrees.
    """

    def __init__(self):
        return super().__init__('{x:g}\xb0')

    def __call__(self, value, pos=None):
        return super().__call__(value * 360, pos=pos)


class PhaseAsRadiansLocator(MaxNLocator):
    """
    Matplotlib locator for phases in the range [0:1] as radians.
    """

    def tick_values(self, vmin, vmax):
        if vmax - vmin > 2.:
            values = super().tick_values(vmin, vmax)
        else:
            power = np.ceil(-np.log2((vmax - vmin) / 3))
            imin = np.floor(vmin / 2**-power)
            imax = np.ceil(vmax / 2**-power)
            values = np.arange(imin, imax + 1) * 2 ** -power
        return values[(values >= vmin) & (values <= vmax)]

    def __call__(self):
        vmin, vmax = self.axis.get_view_interval()
        return self.tick_values(vmin, vmax)


class PhaseAsRadiansFormatter(Formatter):
    """
    Matplotlib formatter for phases in the range [0:1] as radians.
    """

    def __call__(self, value, pos=None):
        value = 2 * value
        fraction = Fraction(value).limit_denominator(1000000)
        top, bot = fraction.numerator, fraction.denominator
        if top < 100 and bot < 100:
            if top == 0:
                return '0'
            elif top == 1 and bot == 1:
                return '\u03c0'
            elif top == -1 and bot == 1:
                return '-\u03c0'
            elif top == 1:
                return '\u03c0/{0}'.format(bot)
            elif top == -1:
                return '-\u03c0/{0}'.format(bot)
            elif bot == 1:
                return '{0}\u03c0'.format(top)
            else:
                return '{0}\u03c0/{1}'.format(top, bot)
        else:
            return '{0:.5g}\u03c0'.format(value)



import os
import glob

import nbformat
from nbconvert.preprocessors import ExecutePreprocessor

for notebook in glob.glob('docs/**/*.ipynb', recursive=True):

    print("Running {0}".format(notebook))

    with open(notebook) as f:
        nb = nbformat.read(f, as_version=4)

    ep = ExecutePreprocessor(timeout=600, kernel_name='python3')
    ep.preprocess(nb, {'metadata': {'path': os.path.dirname(notebook)}})



#!/usr/bin/env python

# NOTE: most of the configuration is defined in setup.cfg

import sys
from distutils.version import LooseVersion

import pip
import setuptools
from setuptools import setup

if LooseVersion(pip.__version__) < '18.0':
    sys.stderr.write("ERROR: pip 18.0 or later is required by aas-timeseries\n")
    sys.exit(1)

if LooseVersion(setuptools.__version__) < '30.3':
    sys.stderr.write("ERROR: setuptools 30.3 or later is required by aas-timeseries\n")
    sys.exit(1)

setup(use_scm_version=True,
      setup_requires=['setuptools_scm'])



import uuid
from collections import OrderedDict

import numpy as np

from astropy.time import Time, TimeDelta
from astropy import units as u
from astropy.units import Quantity
from aas_timeseries.data import Data
from aas_timeseries.layers import BaseLayer, Markers, Line, VerticalLine, VerticalRange, HorizontalLine, HorizontalRange, Range, Text, time_to_vega

__all__ = ['BaseView', 'View']

VALID_TIME_FORMATS = {}
VALID_TIME_FORMATS['absolute'] = ['jd', 'mjd', 'unix', 'iso', 'auto']
VALID_TIME_FORMATS['relative'] = ['seconds', 'hours', 'days', 'years']
VALID_TIME_FORMATS['phase'] = ['unity', 'degrees', 'radians']

VALID_TIME_MODES = ['absolute', 'relative', 'phase']


class BaseView:
    """
    Base class for view-like objects (both the base figure and the actual views)
    """

    def __init__(self, time_mode=None):
        self.uuid = str(uuid.uuid4())
        self._data = OrderedDict()
        self._layers = OrderedDict()
        self._xlim = None
        self._ylim = None
        self._ylog = False
        self._xlabel = ''
        self._ylabel = ''
        self._time_format = ''

        if time_mode is not None and time_mode not in VALID_TIME_MODES:
            raise ValueError("time_mode should be one of " + "/".join(VALID_TIME_MODES))

        self._time_mode = time_mode or 'absolute'

    @property
    def time_mode(self):
        return self._time_mode

    @property
    def ylog(self):
        """
        Whether the y axis is linear (`False`) or log (`True`).
        """
        return self._ylog

    @ylog.setter
    def ylog(self, value):
        self._ylog = value

    @property
    def xlabel(self):
        """
        The label to use for the x-axis. If not specified, this is determined
        automatically from the type of x-axis.
        """
        if self._xlabel:
            return self._xlabel
        else:
            if self._time_mode == 'absolute':
                return 'Time'
            elif self._time_mode == 'relative':
                return 'Relative Time ({0})'.format(self.time_format[0])
            elif self._time_mode == 'phase':
                return 'Phase'

    @xlabel.setter
    def xlabel(self, value):
        self._xlabel = value

    @property
    def ylabel(self):
        """
        The label to use for the y-axis.
        """
        return self._ylabel

    @ylabel.setter
    def ylabel(self, value):
        self._ylabel = value

    @property
    def xlim(self):
        """
        The x/time limits of the view.
        """
        return self._xlim

    @xlim.setter
    def xlim(self, range):
        if not isinstance(range, (tuple, list)) or len(range) != 2:
            raise ValueError("xlim should be a tuple of two elements")
        start, end = range
        if isinstance(start, str):
            start = Time(start)
        if isinstance(end, str):
            end = Time(end)
        if not isinstance(start, Time) or not isinstance(end, Time):
            raise TypeError("xlim should be a typle of two Time instances")
        self._xlim = start, end

    @property
    def ylim(self):
        """
        The y limits of the view.
        """
        return self._ylim

    @ylim.setter
    def ylim(self, range):
        if isinstance(range[0], u.Quantity) is isinstance(range[1], u.Quantity):
            if isinstance(range[0], u.Quantity):
                if not range[0].unit.is_equivalent(range[1].unit):
                    raise u.UnitsError(f'The units of ymin ({range[0].unit}) are '
                                       f'not compatible with the units of ymax ({range[1].unit})')
        else:
            raise ValueError('Either both or neither limit has to be specified '
                             'as a Quantity')
        self._ylim = range

    @property
    def time_format(self):
        """
        The format to use for the x-axis.
        """
        if self._time_format:
            return self._time_format
        else:
            if self._time_mode == 'absolute':
                return 'auto'
            elif self._time_mode == 'relative':
                return 'seconds'
            elif self._time_mode == 'phase':
                return 'unity'

    @time_format.setter
    def time_format(self, value):
        if value in VALID_TIME_FORMATS[self._time_mode]:
            self._time_format = value
        else:
            raise ValueError('time_format should be one of ' + '/'.join(VALID_TIME_FORMATS[self._time_mode]))

    def _validate_time_column(self, time_series, time_column):
        column = time_series[time_column]
        if self._time_mode == 'absolute':
            if not isinstance(column, Time) or isinstance(column, TimeDelta):
                raise TypeError('When in absolute time mode, the time column should be a Time object')
        elif self._time_mode == 'relative':
            if not isinstance(column, (TimeDelta, Quantity)) or (isinstance(column, Quantity) and column.unit.physical_type != 'time'):
                raise TypeError('When in relative time model, the time column should be a TimeDelta or Quantity object with time units')
        elif self._time_mode == 'phase':
            if not isinstance(column, np.ndarray) or (isinstance(column, Quantity) and column.unit.physical_type != 'dimensionless'):
                raise TypeError('When in relative time model, the time column should be a plain array or a dimensionless Quantity')
        else:
            raise ValueError('time_mode should be one of absolute/relative/phase')

    def add_markers(self, *, time_series=None, column=None, time_column='time', **kwargs):
        """
        Add markers, optionally with errorbars, to the figure.

        Parameters
        ----------
        data : `~astropy.timeseries.TimeSeries`
            The time series object containing the data.
        column : str
            The field in the time series containing the data.
        time_column : str, optional
            The field to use for the time on the x-axis.
        error : str, optional
            The field in the time series containing the data uncertainties.
        shape : {'circle', 'square', 'cross', 'diamond', 'triangle-up', 'triangle-down', 'triangle-right', 'triangle-left'}, optional
            The symbol shape.
        size : float or int, optional
            The area in pixels of the bounding box of the symbols.  Note that this
            value sets the area of the symbol; the side lengths will increase with
            the square root of this value.
        color : str or tuple, optional
            The fill color of the symbols.
        opacity : float or int, optional
            The opacity of the fill color from 0 (transparent) to 1 (opaque).
        edge_color : str or tuple, optional
            The edge color of the symbol.
        edge_opacity : float or int, optional
            The opacity of the edge color from 0 (transparent) to 1 (opaque).
        edge_width : float or int, optional
            The thickness of the edge, in pixels.
        label : str, optional
            The label to use to designate the layer in the legend.

        Returns
        -------
        layer : `~aas_timeseries.layers.Markers`
        """
        self._validate_time_column(time_series, time_column)
        if id(time_series) not in self._data:
            self._data[id(time_series)] = Data(time_series)
        markers = Markers(parent=self, data=self._data[id(time_series)], **kwargs)
        # Note that we need to set the column after the data so that the
        # validation works.
        markers.column = column
        markers.time_column = time_column
        self._layers[markers] = {'visible': True}
        return markers

    def add_line(self, *, time_series=None, column=None, time_column='time', **kwargs):
        """
        Add a line to the figure.

        Parameters
        ----------
        data : `~astropy.timeseries.TimeSeries`
            The time series object containing the data.
        column : str
            The field in the time series containing the data.
        time_column : str, optional
            The field to use for the time on the x-axis.
        width : float or int, optional
            The width of the line, in pixels.
        color : str or tuple, optional
            The color of the line.
        opacity : float or int, optional
            The opacity of the line from 0 (transparent) to 1 (opaque).
        label : str, optional
            The label to use to designate the layer in the legend.

        Returns
        -------
        layer : `~aas_timeseries.layers.Line`
        """
        self._validate_time_column(time_series, time_column)
        if id(time_series) not in self._data:
            self._data[id(time_series)] = Data(time_series)
        line = Line(parent=self, data=self._data[id(time_series)], **kwargs)
        # Note that we need to set the column after the data so that the
        # validation works.
        line.column = column
        line.time_column = time_column
        self._layers[line] = {'visible': True}
        return line

    def add_range(self, *, time_series=None, column_lower=None, column_upper=None, time_column='time', **kwargs):
        """
        Add a time dependent range to the figure.

        Parameters
        ----------
        data : `~astropy.timeseries.TimeSeries`
            The time series object containing the data.
        column_lower : str
            The field in the time series containing the lower value of the data
            range.
        column_upper : str
            The field in the time series containing the upper value of the data
            range.
        time_column : str, optional
            The field to use for the time on the x-axis.
        color : str or tuple, optional
            The fill color of the range.
        opacity : float or int, optional
            The opacity of the fill color from 0 (transparent) to 1 (opaque).
        edge_color : str or tuple, optional
            The edge color of the range.
        edge_opacity : float or int, optional
            The opacity of the edge color from 0 (transparent) to 1 (opaque).
        edge_width : float or int, optional
            The thickness of the edge, in pixels.
        label : str, optional
            The label to use to designate the layer in the legend.

        Returns
        -------
        layer : `~aas_timeseries.layers.Range`
        """
        self._validate_time_column(time_series, time_column)
        if id(time_series) not in self._data:
            self._data[id(time_series)] = Data(time_series)
        range = Range(parent=self, data=self._data[id(time_series)], **kwargs)
        # Note that we need to set the columns after the data so that the
        # validation works.
        range.column_lower = column_lower
        range.column_upper = column_upper
        range.time_column = time_column
        self._layers[range] = {'visible': True}
        return range

    def add_vertical_line(self, time, **kwargs):
        """
        Add a vertical line to the figure.

        Parameters
        ----------
        time : `~astropy.time.Time`
            The date/time at which the vertical line is shown.
        width : float or int, optional
            The width of the line, in pixels.
        color : str or tuple, optional
            The color of the line.
        opacity : float or int, optional
            The opacity of the line from 0 (transparent) to 1 (opaque).
        label : str, optional
            The label to use to designate the layer in the legend.

        Returns
        -------
        layer : `~aas_timeseries.layers.VerticalLine`
        """
        line = VerticalLine(parent=self, time=time, **kwargs)
        self._layers[line] = {'visible': True}
        return line

    def add_vertical_range(self, time_lower, time_upper, **kwargs):
        """
        Add a vertical range to the figure.

        Parameters
        ----------
        time_lower : `~astropy.time.Time`
            The date/time at which the range starts.
        time_upper : `~astropy.time.Time`
            The date/time at which the range ends.
        color : str or tuple, optional
            The fill color of the range.
        opacity : float or int, optional
            The opacity of the fill color from 0 (transparent) to 1 (opaque).
        edge_color : str or tuple, optional
            The edge color of the range.
        edge_opacity : float or int, optional
            The opacity of the edge color from 0 (transparent) to 1 (opaque).
        edge_width : float or int, optional
            The thickness of the edge, in pixels.
        label : str, optional
            The label to use to designate the layer in the legend.

        Returns
        -------
        layer : `~aas_timeseries.layers.VerticalRange`
        """
        range = VerticalRange(parent=self, time_lower=time_lower, time_upper=time_upper, **kwargs)
        self._layers[range] = {'visible': True}
        return range

    def add_horizontal_line(self, value, **kwargs):
        """
        Add a horizontal line to the figure.

        Parameters
        ----------
        value : float or int
            The y value at which the horizontal line is shown.
        width : float or int, optional
            The width of the line, in pixels.
        color : str or tuple, optional
            The color of the line.
        opacity : float or int, optional
            The opacity of the line from 0 (transparent) to 1 (opaque).
        label : str, optional
            The label to use to designate the layer in the legend.

        Returns
        -------
        layer : `~aas_timeseries.layers.HorizontalLine`
        """
        line = HorizontalLine(parent=self, value=value, **kwargs)
        self._layers[line] = {'visible': True}
        return line

    def add_horizontal_range(self, value_lower, value_upper, **kwargs):
        """
        Add a horizontal range to the figure.

        Parameters
        ----------
        value_lower : float or int
            The value at which the range starts.
        value_upper : float or int
            The value at which the range ends.
        color : str or tuple, optional
            The fill color of the range.
        opacity : float or int, optional
            The opacity of the fill color from 0 (transparent) to 1 (opaque).
        edge_color : str or tuple, optional
            The edge color of the range.
        edge_opacity : float or int, optional
            The opacity of the edge color from 0 (transparent) to 1 (opaque).
        edge_width : float or int, optional
            The thickness of the edge, in pixels.
        label : str, optional
            The label to use to designate the layer in the legend.

        Returns
        -------
        layer : `~aas_timeseries.layers.HorizontalRange`
        """
        range = HorizontalRange(parent=self, value_lower=value_lower, value_upper=value_upper, **kwargs)
        self._layers[range] = {'visible': True}
        return range

    def add_text(self, **kwargs):
        """
        Add a text label to the figure.

        Parameters
        ----------
        time : `~astropy.time.Time`
            The date/time at which the text is shown.
        value : float or int
            The y value at which the text is shown.
        weight : {'normal', 'bold'}, optional
            The weight of the text.
        baseline : {'alphabetic', 'top', 'middle', 'bottom'}, optional
            The vertical text baseline.
        align : {'left', 'center', 'right'}, optional
            The horizontal text alignment.
        angle : float or int, optional
            The rotation angle of the text in degrees (default 0).
        text : str, optional
            The text label to show.
        color : str or tuple, optional
            The color of the text.
        opacity : float or int, optional
            The opacity of the text from 0 (transparent) to 1 (opaque).
        label : str, optional
            The label to use to designate the layer in the legend.

        Returns
        -------
        layer : `~aas_timeseries.layers.Text`
        """
        text = Text(parent=self, **kwargs)
        self._layers[text] = {'visible': True}
        return text

    @property
    def layers(self):
        return list(self._layers)

    def _get_domains(self, yunit, as_vega=True):

        if self.xlim is None or self.ylim is None:

            all_times = []
            all_values = []

            # If there are symbol layers, we just use those to determine limits
            if any(isinstance(layer, Markers) for layer in self.layers):
                layer_types = (Markers,)
            else:
                layer_types = (Range, Line)

            for layer in self.layers:
                if isinstance(layer, layer_types):

                    if self._time_mode == 'absolute':
                        all_times.append(np.min(layer.data.time_series[layer.time_column]))
                        all_times.append(np.max(layer.data.time_series[layer.time_column]))
                    elif self._time_mode == 'relative':
                        all_times.append(np.nanmin(layer.data.column_to_values(layer.time_column, u.s)))
                        all_times.append(np.nanmax(layer.data.column_to_values(layer.time_column, u.s)))
                    elif self._time_mode == 'phase':
                        all_times.append(np.nanmin(layer.data.column_to_values(layer.time_column, u.one)))
                        all_times.append(np.nanmax(layer.data.column_to_values(layer.time_column, u.one)))

                    all_values.append(np.nanmin(layer.data.column_to_values(layer.column, yunit)))
                    all_values.append(np.nanmax(layer.data.column_to_values(layer.column, yunit)))

            if len(all_times) > 0:
                xlim_auto = np.min(all_times), np.max(all_times)
            else:
                xlim_auto = None

            if len(all_values) > 0:
                ylim_auto = float(np.min(all_values)), float(np.max(all_values))
            else:
                ylim_auto = None

        if self.xlim is None:
            xlim = xlim_auto
        else:
            xlim = self.xlim

        if self.ylim is None:
            ylim = ylim_auto
        else:
            ylim = self.ylim
            if isinstance(ylim[0], u.Quantity):
                ylim = ylim[0].to_value(yunit), ylim[1].to_value(yunit)
            elif yunit is not u.one:
                raise u.UnitsError(f'Limits for y axis are dimensionless but '
                                   f'expected units of {yunit}')

        xlim = xlim_auto if xlim is None else xlim
        ylim = ylim_auto if ylim is None else ylim

        if xlim is not None:
            if self._time_mode == 'absolute' and as_vega:
                x_domain = ({'signal': time_to_vega(xlim[0])},
                            {'signal': time_to_vega(xlim[1])})
            else:
                x_domain = list(xlim)
        else:
            x_domain = None

        if ylim is not None:
            y_domain = list(ylim)
        else:
            y_domain = None

        return x_domain, y_domain


class View(BaseView):

    def __init__(self, figure=None, inherited_layers=None, time_mode=None):
        super().__init__(time_mode=time_mode)
        self._figure = figure
        self._inherited_layers = inherited_layers or OrderedDict()
        self._data = figure._data
        self.ylabel = figure.ylabel

    def show(self, layers):
        self._set_visible(layers, True)

    def hide(self, layers):
        self._set_visible(layers, False)

    def _set_visible(self, layers, visible):
        if isinstance(layers, BaseLayer):
            layers = [layers]
        for layer in layers:
            if layer in self._layers:
                self._layers[layer]['visible'] = visible
            elif layer in self._inherited_layers:
                self._inherited_layers[layer]['visible'] = visible
            else:
                raise ValueError(f'Layer {layer} not in view')

    def remove(self, layer):
        """
        Remove a layer from the view.
        """
        if layer in self._inherited_layers:
            self._inherited_layers.pop(layer)
        elif layer in self._layers:
            self._layers.pop(layer)
        else:
            raise ValueError(f"Layer '{layer.label}' is not in view")

    @property
    def layers(self):
        return list(self._inherited_layers) + list(self._layers)



import os

from astropy import units as u
from astropy.timeseries import TimeSeries

from aas_timeseries.visualization import InteractiveTimeSeriesFigure
from aas_timeseries.screenshot import interactive_screenshot


def test_interactive_screenshot(tmpdir):

    ts = TimeSeries(time_start='2016-03-22T12:30:31',
                    time_delta=3 * u.s, n_samples=5)
    ts['flux'] = [1, 2, 3, 4, 5]

    filename_json = tmpdir.join('figure.json').strpath
    filename_png = tmpdir.join('figure').strpath

    figure = InteractiveTimeSeriesFigure()
    markers = figure.add_markers(time_series=ts, column='flux', label='Markers')
    figure.add_line(time_series=ts, column='flux', label='Line')

    figure.add_view(title="only markers", include=[markers])

    figure.save_vega_json(filename_json, embed_data=True)

    interactive_screenshot(filename_json, filename_png)

    assert os.path.exists(filename_png + '.png')
    assert os.path.exists(filename_png + '_view1.png')
